import logging
import functools
from tornado import httpclient
from tornado.httpclient import HTTPRequest, HTTPResponse
from ..downloader.base import BaseDownloader
from .. import setting

logger = logging.getLogger(__name__)


class AsyncDownloader(BaseDownloader):
    """
    parallel downloader.
    """

    middleware = ()

    client = httpclient.AsyncHTTPClient(max_clients=setting.max_concurrency)

    def fetch(self, request: HTTPRequest) -> None:
        """
        fetch request and call callback with response
        """

        request = functools.reduce(lambda item, func: None if not item else func(item),
                                   (request, *self.middleware))

        if not request:
            logger.info('page {0} filtered.'.format(request.url))
            return None

        self.client.fetch(request, self.handle)

    def handle(self, response: HTTPResponse):
        """
        handle response and middleware, call callback func
        :param response: HTTPResponse object
        :return: None
        """

        response = functools.reduce(lambda item, func: None if not item else func(item),
                                    (response, *reversed(self.middleware)))

        callback = getattr(response.request, 'callback')
        callback(response)



import logging
from tornado.httpclient import HTTPRequest, HTTPResponse
from ..downloader.base import BaseDownloader
from ..downloader.async import AsyncDownloader
from .. import setting

logger = logging.getLogger(__name__)


class BaseScheduler(object):
    """
    Interface for Scheduler class
    """

    def __init__(self, downloader: BaseDownloader):
        self.downloader = downloader if downloader else AsyncDownloader()
        self.max_concurrency = setting.max_concurrency
        self.concurrency = 0

    def put(self, request: HTTPRequest) -> None:
        """
        put request item from scheduler
        :param request: HTTPRequest object
        :return: None
        """
        pass

    def get(self) -> HTTPRequest:
        """
        get request obj and callback func from scheduler
        :return: HTTPRequest object
        """
        pass

    def fetch(self, request: HTTPRequest):
        """
        wrapper for downloader's fetch method.
        :param request: HTTPRequest object.
        :return: None
        """

        self.concurrency += 1

        setattr(request, 'callback', self.handle)
        self.downloader.fetch(request)

    def handle(self, response: HTTPResponse):
        """
        wrapper for downloader's handle method.
        :param response: HTTPResponse object
        :return: None
        """

        self.concurrency -= 1

        # try maximize downloader's concurrency
        while not self.empty() and self.concurrency < self.max_concurrency:
            self.fetch(self.get())

        # retry if fetch error
        if response.code == 599:
            retry_times = getattr(response.request, 'retry_times', 0)

            if retry_times < setting.max_retry:
                logger.warning('page {0} fetch failed, err: {1}, retry... ({2}/{3})'
                               .format(response.request.url, response.error, retry_times + 1, setting.max_retry))
                setattr(response.request, 'retry_times', retry_times + 1)
                self.put(response.request)
                return None
            else:
                logger.error("page {0} fetch failed. max_retry times tried.".format(response.request.url))
                return None

        logger.info('page {0} fetch success.'.format(response.request.url))
        parser = getattr(response.request, 'parser')
        parser(response)

    def empty(self) -> bool:
        """
        scheduler's queue empty or not
        :return: Boolean, True or False
        """
        pass



from tornado.httpclient import HTTPRequest


class BaseDownloader(object):
    """
    Interface for Downloader
    """

    middleware = ()

    def fetch(self, request: HTTPRequest) -> None:
        """
        Load middleware and fetch page.
        You need to implement fetch by extends this func.
        :param request: HTTPRequest object
        :return: None
        """
        pass



from typing import Generator
from tornado.ioloop import IOLoop
from tornado.httpclient import HTTPRequest, HTTPResponse
from .. import setting


class BaseSpider(object):
    """
    Interface for spider.
    """

    scheduler = None

    def start(self) -> Generator:
        """
        Spider start function. Call once on spider start.
        :return: a generator of tuple, HTTPRequest object and callback func
        """
        pass

    def parse(self, response: HTTPResponse) -> Generator:
        """
        Default parse function.
        :param response:
        :return:
        """
        pass

    def run(self):
        """
        run spider
        :return: None
        """

        for request in self.start():
            assert isinstance(request, HTTPRequest)

            if not getattr(request, 'parser', None):
                setattr(request, 'parser', self.parse)

            self.scheduler.put(request)

        loop = IOLoop.current(instance=True)
        try:
            loop.start()
        except KeyboardInterrupt:
            loop.stop()



from queue import Queue
from tornado.httpclient import HTTPRequest
from ..downloader.base import BaseDownloader
from ..scheduler.base import BaseScheduler


class QueueScheduler(BaseScheduler):
    """
    In memory queue scheduler.
    Be careful of memory run out.
    """

    def __init__(self, downloader: BaseDownloader = None):
        super(QueueScheduler, self).__init__(downloader)
        self.queue = Queue()

    def get(self):
        if not self.queue.empty():
            return self.queue.get_nowait()
        else:
            return None

    def put(self, request: HTTPRequest):

        if self.concurrency < self.max_concurrency:
            self.fetch(request)
        else:
            self.queue.put_nowait(request)

    def empty(self):
        return self.queue.empty()
